from typing import Callable, Awaitable

from aio_pika import Message
from aio_pika.abc import ConsumerTag, AbstractRobustChannel, AbstractRobustQueue, AbstractIncomingMessage, AbstractRobustConnection

from ploomby.abc.exceptions import UnregisteredHandler, NoConnectionError, NoMessageKeyError
from ploomby.abc import IncomingMessageHandler, RawDataHandler, MessageKeyType, HandlerDependencyType
from ploomby.logger import logger


class RabbitConsumer:
    """
    Simplest consumer for RabbitMQ implemented by aio_pika.  
    """

    def __init__(
            self,
            connection_dep: Callable[[], Awaitable[AbstractRobustConnection]],
            message_key_name: str,
            conn_is_shared: bool = True,
            prefetch_count: int = 0,
            reconnect: bool = True,
    ):
        """
        :param conn_url: URL for connection with RbbitMQ
        :type conn_url: str
        :param message_key_name: The key is in the header dictionary, which can be used to uniquely identify the incoming message. For example if message_key_name is 'task_name' that means consumer will try to get value from message headers by key 'task_name' and use this value as key to get handler from handlers map
        :type message_key_name: str
        :param prefetch_count: Number of unacknowledged messages that consumer can keep at the same time
        :type prefetch_count: int
        :param reconnect: If false an exception will be raised when connection will become lost
        :type reconnect: bool
        """
        self._connection_dep = connection_dep
        self._connection: AbstractRobustConnection = None
        self._conn_is_shared = conn_is_shared
        self._queue: AbstractRobustQueue = None
        self._tag: ConsumerTag = None
        self._channel: AbstractRobustChannel = None
        self._prefetch_count = prefetch_count
        self._reconnect = reconnect
        self.message_key_name = message_key_name
        self._handlers_map: dict[MessageKeyType, RawDataHandler] = {}

    async def _cancel_queue(self):
        if self._queue:
            await self._queue.cancel(self._tag)
            self._tag = None
            self._queue = None

    async def _close_channel(self):
        if self._channel:
            if not self._channel.is_closed:
                await self._channel.close()
            self._channel = None

    async def _close_conn(self):
        if self._connection:
            if not self._connection.is_closed:
                if not self._conn_is_shared:
                    await self._connection.close()
            self._connection = None

    async def connect(self):
        self._connection = await self._connection_dep()

    async def disconnect(self) -> None:
        await self._cancel_queue()
        await self._close_channel()
        await self._close_conn()

    async def _check_connection(self):
        if not self._connection or self._connection.is_closed:
            if self._reconnect:
                await self.connect()
            else:
                raise NoConnectionError("Connection lost")

    async def _init_channel(self):
        await self._check_connection()
        self._channel = await self._connection.channel(publisher_confirms=False)

    async def _get_channel(self):
        if not self._channel or self._channel.is_closed:
            await self._init_channel()
        return self._channel

    async def _declare_queue(self, queue_name: str):
        channel = await self._get_channel()
        queue = await channel.declare_queue(
            queue_name,
            durable=True,
            arguments={
                "x-max-priority": 10,
            },
        )
        return queue

    def _on_message(self, get_handler_func: HandlerDependencyType) -> IncomingMessageHandler:
        async def handle_message(message: AbstractIncomingMessage) -> None:
            publisher_answer = b"ACK"
            try:
                message_key = message.headers.get(self.message_key_name)
                if not message_key:
                    raise NoMessageKeyError(
                        f"Headers do not contain value by key '{self.message_key_name}'")
                handler = get_handler_func(message_key)
                if not handler:
                    raise UnregisteredHandler(
                        f"Handler with key '{self.message_key_name}' was not registered")
                await handler(message.body.decode())
            except Exception as e:
                logger.error(
                    f"Error occured when handled message retrieved from '{self._queue.name}': {e}")
                publisher_answer = b"NACK"
            await message.ack()
            if message.reply_to:
                channel = await self._get_channel()
                await channel.default_exchange.publish(
                    Message(body=publisher_answer), routing_key=message.reply_to
                )
        return handle_message

    async def consume(self, listen_for: str, get_handler_func: HandlerDependencyType):
        self._queue = await self._declare_queue(listen_for)
        self._tag = await self._queue.consume(self._on_message(get_handler_func))



from typing import Callable, Awaitable

from aio_pika.abc import AbstractRobustConnection
from aio_pika import connect_robust

from ploomby.abc import MessageConsumer

from .consumer import RabbitConsumer


class RabbitConsumerFactory:
    def __init__(self, conn_url: str, shared_conn: bool = True):
        self._conn_url = conn_url
        self._shared_conn = shared_conn
        self._connection: AbstractRobustConnection = None

    async def _get_connection(self) -> AbstractRobustConnection:
        if self._shared_conn:
            if not self._connection or self._connection.is_closed:
                self._connection = await connect_robust(self._conn_url)
            return self._connection
        return await connect_robust(self._conn_url)

    async def create(self, message_key_name: str, prefetch_count: int = 1, reconnect: bool = True) -> MessageConsumer:
        consumer = RabbitConsumer(
            self._get_connection,
            message_key_name,
            conn_is_shared=self._shared_conn,
            prefetch_count=prefetch_count,
            reconnect=reconnect,
        )
        await consumer.connect()
        return consumer



class UnregisteredHandler(Exception):
    pass


class NoConnectionError(Exception):
    pass


class NoModelProvidedError(Exception):
    pass


class NoMessageKeyError(Exception):
    pass


class ConsumerAlreadyRegistered(Exception):
    pass



from typing import Type, Awaitable, Any
from functools import wraps

from pydantic import BaseModel

from ploomby.abc import (
    MessageKeyType,
    HandlerType,
    MessageConsumer,
    MessageConsumerFactory,
    NoModelProvidedError,
    ConsumerAlreadyRegistered
)
from ploomby.logger import logger

__all__ = [
    "HandlersRegistry",
    "MessageConsumerRegistry"
]


class HandlersRegistry:
    def __init__(self):
        self._handlers: dict[MessageKeyType, HandlerType] = {}
        self._models: dict[MessageKeyType, Type[BaseModel]] = {}

    def _find_model(self, annotations: dict):
        for arg in annotations.values():
            if issubclass(arg, BaseModel):
                return arg
        return None

    def register(self, key: MessageKeyType):
        def decorator(handler_func: HandlerType):
            model = self._find_model(handler_func.__annotations__)
            if not model:
                raise NoModelProvidedError(
                    f"No data model provided into handler '{handler_func.__name__}'")
            self._handlers[key] = handler_func
            self._models[key] = model
            logger.info(
                f"Handler '{handler_func.__name__}' registered by key '{key}'. Handler excpects data as model of {model.__name__}")

            @wraps(handler_func)
            def _(dto: BaseModel, *args, **kwargs):
                pass
            return _
        return decorator

    def generate_handler_with_validation_coro(self, handler_key: MessageKeyType):
        def handler_with_validation(data: str | bytes) -> Awaitable[Any]:
            model = self._models.get(handler_key)
            handler = self._handlers.get(handler_key)
            if not (model and handler):
                return None
            dto = model.model_validate_json(data)
            return handler(dto)
        return handler_with_validation


class MessageConsumerRegistry:
    def __init__(
            self,
            handlers_registry: HandlersRegistry,
            consumer_factory: MessageConsumerFactory
    ):
        self._handlers_registry = handlers_registry
        self._consumers_map: dict[str, MessageConsumer] = {}
        self._consumer_factory = consumer_factory

    async def register(
            self,
            listen_for: str,
            message_key_name: str,
            *args,
            **kwargs
    ):
        """
        Uses provided factory to create consumer instance and subscribe it on provided resource.
        If want to use not built-in factories just define it according to required interface of factore and provide to registry

        :param listen_for: Name of representation of what the consumer is subscribed to
        :type listen_for: str
        :param message_key_name: Value that consumer should to use to get value from message headers(or other metadata)
        to identify incoming messsage and get corresponding handler using get_handler_func provided in consume() 
        :type message_key_name: str
        :param args: Args using to provide to create method of consumer factory
        :param kwargs: Kwargs using to provide to create method of consumer factory
        """
        current = self._consumers_map.get(listen_for)
        if current:
            raise ConsumerAlreadyRegistered(f"Consumer for '{listen_for}' already registered")
        consumer = await self._consumer_factory.create(message_key_name, *args, **kwargs)
        await consumer.consume(listen_for, self._handlers_registry.generate_handler_with_validation_coro)
        self._consumers_map[listen_for] = consumer
        logger.info(
            f"Consumer registered with message key name '{consumer.message_key_name}'. Listening for '{listen_for}'")



from typing import Protocol

from .types import HandlerDependencyType


class MessageConsumer(Protocol):
    """
    Interface of consumer that should be implemented to use in registries.
    Message key name is a value that consumer should to use to get value from message headers(or other metadata)
    to identify incoming messsage and get corresponding handler using get_handler_func provided in consume() 
    """
    message_key_name: str

    async def connect(self) -> None: ...
    async def disconnect(self) -> None: ...

    async def consume(self, listen_for: str, get_handler_func: HandlerDependencyType):
        """
        Starts consume events/messages from resourse

        :param listen_for: Name of representation of what the consumer is subscribed to
        :type listen_for: str
        :param get_handler_func: Wrapper that retrieves message key and returns function, that handles raw data from broker
        :type get_handler_func: HandlerDependencyType

        examples get_handler_func::

            def get_order_handler(key: str) -> Callable[[str | bytes], Awaitable[None]]:
                handlers = {
                    "order.created": handle_order_created,
                    "order.updated": handle_order_updated,
                    "order.cancelled": handle_order_cancelled,
                }
                return handlers.get(key.decode("utf-8"), default_handler)

            async def handle_order_created(raw_data: str | bytes) -> None:
                order_data = json.loads(raw_data)
                # Process order creation logic
                await process_new_order(order_data)

            async def handle_order_updated(raw_data: str | bytes) -> None:
                order_data = json.loads(raw_data)
                # Process order update logic
                await update_order(order_data)

            async def default_handler(raw_data: str | bytes) -> None:
                logger.warning(f"Unhandled message key with data: {raw_data}")
        """



import sys
import os

from datetime import timezone

from loguru import logger

logger.remove()


def to_utc(record):
    if not os.getenv("DEBUG"):
        record["time"] = record["time"].astimezone(timezone.utc)
    return True


logger.add(
    sys.stdout,
    colorize=True,
    filter=to_utc
)
